@@ -1,7 +1,7 @@
 # ============================================================
 #
 # Copyright (C) 2010 by Johannes Wienke <jwienke at techfak dot uni-bielefeld dot de>
-# Copyright (C) 2011, 2012 Jan Moringen <jmoringe@techfak.uni-bielefeld.de>
+# Copyright (C) 2011, 2012, 2013 Jan Moringen <jmoringe@techfak.uni-bielefeld.de>
 #
 # This file may be licensed under the terms of the
 # GNU Lesser General Public License Version 3 (the ``LGPL''),
@@ -32,7 +32,12 @@ registering and selecting them.
 @author: plueckin
 """
 
-from numbers import Integral
+from numbers import Integral, Real
+import struct
+from rsb.protocol.collections.EventsByScopeMap_pb2 import EventsByScopeMap
+from rsb.transport.conversion import notificationToEvent, eventToNotification
+from rsb import Scope
+from threading import RLock
 
 class Converter(object):
     """
@@ -97,10 +102,10 @@ class Converter(object):
 
     wireSchema = property(getWireSchema)
 
-    def serialize(self, input):
+    def serialize(self, inp):
         raise NotImplementedError()
 
-    def deserialize(self, input, wireSchema):
+    def deserialize(self, inp, wireSchema):
         raise NotImplementedError()
 
 class UnknownConverterError(KeyError):
@@ -166,14 +171,26 @@ class ConverterMap(ConverterSelectionStrategy):
         self._converters[key] = converter
 
     def _getConverterForWireSchema(self, wireSchema):
-        for ((converterWireSchema, ignored), converter) in self._converters.items():
+        for ((converterWireSchema, _), converter) in self._converters.items():
             if converterWireSchema == wireSchema:
                 return converter
 
     def _getConverterForDataType(self, dataType):
-        for ((ignored, converterDataType), converter) in self._converters.items():
+        # If multiple converters are applicable, use most specific.
+        candidates = []
+        for ((_, converterDataType), converter) in self._converters.items():
             if issubclass(dataType, converterDataType):
-                return converter
+                candidates.append(converter)
+        if len(candidates) == 1:
+            return candidates[0]
+        elif len(candidates) > 1:
+            def compareViaSubclass(x, y):
+                if issubclass(x, y):
+                    return -1
+                else:
+                    return 1
+            return sorted(candidates, compareViaSubclass,
+                          key = lambda x: x.getDataType())[0]
 
     def getConverters(self):
         return self._converters
@@ -206,9 +223,9 @@ class PredicateConverterList (ConverterMap):
                      dataTypePredicate=None,
                      replaceExisting=True):
         if wireSchemaPredicate is None:
-            #if converter.getWireSchema() == 'void':
+            # if converter.getWireSchema() == 'void':
             #    wireSchemaPredicate = lambda wireSchema: True
-            #else:
+            # else:
                 wireSchemaPredicate = lambda wireSchema: wireSchema == converter.getWireSchema()
         if dataTypePredicate is None:
             dataTypePredicate = lambda dataType: dataType == converter.getDataType()
@@ -217,12 +234,12 @@ class PredicateConverterList (ConverterMap):
         self._list.append((key, converter))
 
     def _getConverterForWireSchema(self, wireSchema):
-        for ((predicate, ignored), converter) in self._list:
+        for ((predicate, _), converter) in self._list:
             if predicate(wireSchema):
                 return converter
 
     def _getConverterForDataType(self, dataType):
-        for ((ignored, predicate), converter) in self._list:
+        for ((_, predicate), converter) in self._list:
             if predicate(dataType):
                 return converter
 
@@ -240,6 +257,7 @@ class UnambiguousConverterMap (ConverterMap):
                                        % (converter.getDataType(), wireSchema, dataType))
         super(UnambiguousConverterMap, self).addConverter(converter, replaceExisting)
 
+__globalConverterMapsLock = RLock()
 __globalConverterMaps = {}
 
 def registerGlobalConverter(converter, replaceExisting=False):
@@ -265,9 +283,10 @@ def getGlobalConverterMap(wireType):
     @return: converter map constantly updated
     """
 
-    if not wireType in __globalConverterMaps:
-        __globalConverterMaps[wireType] = ConverterMap(wireType)
-    return __globalConverterMaps[wireType]
+    with __globalConverterMapsLock:
+        if not wireType in __globalConverterMaps:
+            __globalConverterMaps[wireType] = ConverterMap(wireType)
+        return __globalConverterMaps[wireType]
 
 # --- converters with bytearray as serialization type ---
 
@@ -281,10 +300,10 @@ class IdentityConverter (Converter):
     def __init__(self):
         super(IdentityConverter, self).__init__(bytearray, type(None), 'void')
 
-    def serialize(self, input):
+    def serialize(self, inp):
         return bytearray(), self.wireSchema
 
-    def deserialize(self, input, wireSchema):
+    def deserialize(self, inp, wireSchema):
         pass
 
     def AlwaysApplicable(self):
@@ -303,14 +322,66 @@ class NoneConverter (Converter):
     def __init__(self):
         super(NoneConverter, self).__init__(bytearray, type(None), 'void')
 
-    def serialize(self, input):
+    def serialize(self, inp):
         return bytearray(), self.wireSchema
 
-    def deserialize(self, input, wireSchema):
+    def deserialize(self, inp, wireSchema):
         assert wireSchema == self.wireSchema
 
         pass
 
+def makeStructBasedConverter(name, dataType, wireSchema, fmt, size):
+    class NewConverter(Converter):
+        def __init__(self):
+            super(self.__class__, self).__init__(bytearray, dataType, wireSchema)
+
+        def serialize(self, inp):
+            return bytearray(struct.pack(fmt, inp)), self.wireSchema
+
+        def deserialize(self, inp, wireSchema):
+            assert wireSchema == self.wireSchema
+            return struct.unpack(fmt, str(inp))[0]
+
+    NewConverter.__name__ = name
+    # TODO(jmoringe): seems to be impossible in CPython
+    # NewConverter.__doc__ = """
+    # A converter that serializes %(dataType)s to bytearrays with %(wireSchema)s wire-schema.
+    #
+    # @author: jmoringe
+    # """ % {
+    #     "dataType":   dataType,
+    #     "wireSchema": wireSchema
+    # }
+
+    globals()[name] = NewConverter
+    return NewConverter
+
+makeStructBasedConverter('DoubleConverter', Real, 'double', '<d', 8)
+makeStructBasedConverter('FloatConverter', Real, 'float', '<f', 4)
+makeStructBasedConverter('Uint32Converter', Integral, 'uint32', '<I', 4)
+makeStructBasedConverter('Int32Converter', Integral, 'int32', '<i', 4)
+makeStructBasedConverter('Uint64Converter', Integral, 'uint64', '<Q', 8)
+makeStructBasedConverter('Int64Converter', Integral, 'int64', '<q', 8)
+makeStructBasedConverter('BoolConverter', bool, 'bool', '?', 1)
+
+# Registered at end of file
+class BytesConverter(Converter):
+    """
+    Handles byte arrays.
+
+    @author: jmoringe
+    """
+
+    def __init__(self, wireSchema="bytes", dataType=bytearray):
+        super(BytesConverter, self).__init__(bytearray, dataType, wireSchema)
+
+    def serialize(self, inp):
+        return inp, self.wireSchema
+
+    def deserialize(self, inp, wireSchema):
+        assert(wireSchema == self.wireSchema)
+        return inp
+
 class StringConverter(Converter):
     """
     A converter that serializes strings to bytearrays with a specified
@@ -323,42 +394,32 @@ class StringConverter(Converter):
         super(StringConverter, self).__init__(bytearray, dataType, wireSchema)
         self.__encoding = encoding
 
-    def serialize(self, input):
-        return bytearray(input.encode(self.__encoding)), self.wireSchema
+    def serialize(self, inp):
+        return bytearray(inp.encode(self.__encoding)), self.wireSchema
 
-    def deserialize(self, input, wireSchema):
-        type = self.getDataType()
-        if type == unicode:
-            return type(str(input), self.__encoding)
-        elif type == str:
-            return str(input)
+    def deserialize(self, inp, wireSchema):
+        dataType = self.getDataType()
+        if dataType == unicode:
+            return dataType(str(inp), self.__encoding)
+        elif dataType == str:
+            return str(inp)
         else:
             raise ValueError("Inacceptable dataType %s" % type)
 
-class Uint64Converter(Converter):
+class ByteArrayConverter(Converter):
     """
-    A converter that serializes unsigned integers that fit in 64 bits.
+    A converter which just passes through the original byte array of a message.
 
-    @author: jmoringe
+    @author: jwienke
     """
-
     def __init__(self):
-        super(Uint64Converter, self).__init__(bytearray, Integral, 'uint64')
+        super(ByteArrayConverter, self).__init__(bytearray, bytearray, '.*')
 
-    def serialize(self, input):
-        if input < 0 or input > ((1 << 64) - 1):
-            raise ValueError, '%s is invalid as uint64 value' % input
+    def serialize(self, data):
+        return bytearray(data)
 
-        output = bytearray('12345678')
-        for i in range(8):
-            output[i] = (input & (0xff << (i * 8))) >> (i * 8)
-        return output, self.wireSchema
-
-    def deserialize(self, input, wireSchema):
-        output = 0L
-        for i in range(8L):
-            output |= (long(input[i]) << (i * 8L))
-        return output
+    def deserialize(self, data, wireSchema):
+        return bytearray(data)
 
 class ProtocolBufferConverter(Converter):
     """
@@ -385,16 +446,16 @@ class ProtocolBufferConverter(Converter):
     def getMessageClassName(self):
         return self.messageClass.DESCRIPTOR.full_name
 
-    def serialize(self, input):
-        return bytearray(input.SerializeToString()), self.wireSchema
+    def serialize(self, inp):
+        return bytearray(inp.SerializeToString()), self.wireSchema
 
-    def deserialize(self, input, wireSchema):
+    def deserialize(self, inp, wireSchema):
         assert wireSchema == self.wireSchema
 
         output = self.messageClass()
         # we need to convert back to string because bytearrays do not work with
         # protobuf
-        output.ParseFromString(str(input))
+        output.ParseFromString(str(inp))
         return output
 
     def __str__(self):
@@ -404,6 +465,65 @@ class ProtocolBufferConverter(Converter):
     def __repr__(self):
         return str(self)
 
+class EventsByScopeMapConverter(Converter):
+    """
+    A converter for aggregated events ordered by their scope and time for each
+    scope. As a client data type dictionaries are used. Think about this when
+    you register the converter and also have other dictionaries to transmit.
+
+    @author: jwienke
+    """
+
+    def __init__(self, converterRepository=getGlobalConverterMap(bytearray)):
+        self.__converterRepository = converterRepository
+        self.__converter = ProtocolBufferConverter(EventsByScopeMap)
+        super(EventsByScopeMapConverter, self).__init__(bytearray, dict, self.__converter.wireSchema)
+
+    def serialize(self, data):
+
+        eventMap = EventsByScopeMap()
+
+        for scope, events in data.iteritems():
+
+            scopeSet = eventMap.sets.add()
+            scopeSet.scope = scope.toString()
+
+            for event in events:
+
+                wire, wireSchema = self.__converterRepository.getConverterForDataType(type(event.data)).serialize(event.data)
+
+                notification = scopeSet.notifications.add()
+                eventToNotification(notification, event, wireSchema, wire, True)
+
+        return self.__converter.serialize(eventMap)
+
+    def deserialize(self, wire, wireSchema):
+        preliminaryMap = self.__converter.deserialize(wire, wireSchema)
+
+        output = {}
+
+        for scopeSet in preliminaryMap.sets:
+            scope = Scope(scopeSet.scope)
+            output[scope] = []
+            for notification in scopeSet.notifications:
+
+                converter = self.__converterRepository.getConverterForWireSchema(notification.wire_schema)
+                event = notificationToEvent(notification, notification.data, notification.wire_schema, converter)
+
+                output[scope].append(event)
+
+        return output
+
+# FIXME We do not register all available converters here to avoid
+# ambiguities.
 registerGlobalConverter(NoneConverter())
+registerGlobalConverter(DoubleConverter())
+#registerGlobalConverter(FloatConverter())
+#registerGlobalConverter(Uint32Converter())
+#registerGlobalConverter(Int32Converter())
+#registerGlobalConverter(Uint64Converter())
+registerGlobalConverter(Int64Converter())
+registerGlobalConverter(BoolConverter())
+registerGlobalConverter(BytesConverter())
 registerGlobalConverter(StringConverter(wireSchema="utf-8-string", dataType=str, encoding="utf_8"))
-registerGlobalConverter(Uint64Converter())
+registerGlobalConverter(ByteArrayConverter())
